#include "tpl.h"

#include "term.h"

/** @brief tpl packing format for a term.
 *
 * The pack elements are: 1. term type (char); 2. data (string); 3. void* type
 * metadata, cast to 8-byte unsigned.
 */
#define TERM_PACK_FMT "csU"

#define MAX_VALID_TERM_TYPE     LSUP_TERM_BNODE /* For type validation. */


/*
 * Data structures.
 */

struct iri_info_t {
    LSUP_NSMap *        nsm;        // NSM handle for prefixed IRI.
    regmatch_t          prefix;     // Matching group #1.
    regmatch_t          path;       // Matching group #5.
    regmatch_t          frag;       // Matching group #10.
};


/*
 * Extern variables.
 */

struct term_cache_t *LSUP_term_cache = NULL;
uint32_t LSUP_default_dtype_key = 0;
regex_t *LSUP_uri_ptn;
LSUP_Term *LSUP_default_datatype = NULL;


/*
 * Static variables.
 */

// Characters not allowed in a URI string.
static const char *invalid_uri_chars = "<>\" {}|\\^`";


/*
 * Static prototypes.
 */

LSUP_rc
term_init(
        LSUP_Term *term, LSUP_TermType type, const char *data, void *metadata);


/*
 * Term API.
 */

LSUP_Term *
LSUP_term_new (
        LSUP_TermType type, const char *data, void *metadata)
{
    LSUP_Term *term;
    CALLOC_GUARD (term, NULL);

    // If undefined, just set the type.
    if (type == LSUP_TERM_UNDEFINED) term->type = type;

    else if (UNLIKELY (term_init (
                    term, type, data, metadata) != LSUP_OK)) {
        free (term);
        return NULL;
    }

    return term;
}


LSUP_Term *
LSUP_term_new_from_buffer (const LSUP_Buffer *sterm)
{
    if (UNLIKELY (!sterm)) return NULL;

    LSUP_Term *term = NULL;
    LSUP_TermType type;
    char *data = NULL;
    void *metadata;

    tpl_node *tn;

    tn = tpl_map (TERM_PACK_FMT, &type, &data, &metadata);
    if (UNLIKELY (!tn)) goto finally;

    if (UNLIKELY (tpl_load (tn, TPL_MEM, sterm->addr, sterm->size) < 0)) {
        log_error ("Error loading serialized term.");
        goto finally;
    }
    if (UNLIKELY (tpl_unpack (tn, 0) < 0)) {
        log_error ("Error unpacking serialized term.");
        goto finally;
    }

    if (type == LSUP_TERM_LT_LITERAL)
        term = LSUP_lt_literal_new (data, (char *)&metadata);
    else term = LSUP_term_new (type, data, metadata);

finally:
    tpl_free (tn);
    free (data);

    return term;
}


LSUP_Buffer *
LSUP_term_serialize (const LSUP_Term *term)
{
    /*
     * In serializing a term, the fact that two terms of different types may
     * be semantically identical must be taken into account. Specifically, a
     * namespace-prefixed IRI ref is identical to its fully qualified version,
     * and a LSUP_TERM_LT_LITERAL with no language tag is identical to a
     * LSUP_TERM_LITERAL of xsd:string type, made up of the same string. Such
     * terms must have identical serializations.
     */

    if (UNLIKELY (!term)) return NULL;

    LSUP_Term *tmp_term;
    void *metadata = NULL;

    if (term->type == LSUP_TERM_NS_IRIREF) {
        // For IRI refs, simply serialize the FQ version of the term.
        char *fq_uri;

        if (LSUP_nsmap_normalize_uri (
            term->iri_info->nsm, term->data, &fq_uri
        ) != LSUP_OK) return NULL;

        tmp_term = LSUP_iriref_new (fq_uri, NULL);

    } else if (term->type == LSUP_TERM_LT_LITERAL) {
        // For LT literals with empty lang tag, convert to a normal xsd:string.
        if (strlen (term->lang) == 0)
            tmp_term = LSUP_literal_new (term->data, NULL);
        else tmp_term = LSUP_lt_literal_new (term->data, (char *) term->lang);

    } else tmp_term = LSUP_term_new (
            term->type, term->data, (void *) term->datatype);
    // "datatype" can be anything here since it's cast to void *.

    // metadata field is ignored for IRI ref.
    if (tmp_term->type == LSUP_TERM_LITERAL)
        metadata = tmp_term->datatype;
    else if (tmp_term->type == LSUP_TERM_LT_LITERAL)
        memcpy (&metadata, tmp_term->lang, sizeof (metadata));

    LSUP_Buffer *sterm;
    MALLOC_GUARD (sterm, NULL);

    int rc = tpl_jot (
            TPL_MEM, &sterm->addr, &sterm->size, TERM_PACK_FMT,
            &tmp_term->type, &tmp_term->data, &metadata);
    LSUP_term_free (tmp_term);

    if (rc != 0) {
        LSUP_buffer_free (sterm);
        return NULL;
    }

    return sterm;
}


LSUP_rc
term_init (
        LSUP_Term *term, LSUP_TermType type,
        const char *data, void *metadata)
{
    if (UNLIKELY (!LSUP_uri_ptn)) {
        log_error ("Environment not initialized. Did you call LSUP_init()?");
        return LSUP_ERROR;
    }
    // This can never be LSUP_TERM_UNDEFINED.
    if (type <= LSUP_TERM_UNDEFINED || type > MAX_VALID_TERM_TYPE) {
        log_error ("%d is not a valid term type.", type);
        return LSUP_VALUE_ERR;
    }

    term->type = type;

    if (data) {
        // Validate IRI.
        if (LSUP_IS_IRI (term)) {
            char *fquri;

            // Find fully qualified IRI to parse.
            if (term->type == LSUP_TERM_NS_IRIREF) {
                if (LSUP_nsmap_normalize_uri (
                    metadata, data, &fquri) != LSUP_OK
                ) {
                    log_error ("Error normalizing IRI data.");

                    return LSUP_VALUE_ERR;
                }
                log_debug ("Fully qualified IRI: %s", fquri);
            } else fquri = (char *) data;

            if (strpbrk (fquri, invalid_uri_chars) != NULL) {
                log_error (
                        "Characters %s are not allowed. Got: %s\n",
                        invalid_uri_chars, fquri);

                return LSUP_VALUE_ERR;
            }

            // Capture interesting IRI parts.
            regmatch_t matches[11];
            if (UNLIKELY (regexec (LSUP_uri_ptn, fquri, 11, matches, 0) != 0)) {
                fprintf (stderr, "Error matching URI pattern.\n");

                return LSUP_VALUE_ERR;
            }
            if (term->type == LSUP_TERM_NS_IRIREF) free (fquri);

            MALLOC_GUARD (term->iri_info, LSUP_MEM_ERR);

            term->iri_info->prefix = matches[1];
            term->iri_info->path = matches[5];
            term->iri_info->frag = matches[10];
            term->iri_info->nsm = metadata;
        }

        term->data = strdup (data);

    } else {
        // No data. Make up a random UUID or URI if allowed.
        if (type == LSUP_TERM_IRIREF || type == LSUP_TERM_BNODE) {
            uuid_t uuid;
            uuid_generate_random (uuid);

            uuid_str_t uuid_str;
            uuid_unparse_lower (uuid, uuid_str);

            if (type == LSUP_TERM_IRIREF) {
                term->data = malloc (UUID4_URN_SIZE);
                snprintf (
                        term->data, UUID4_URN_SIZE, "urn:uuid4:%s", uuid_str);

                MALLOC_GUARD (term->iri_info, LSUP_MEM_ERR);

                // Allocate IRI match patterns manually.
                term->iri_info->prefix.rm_so = 0;
                term->iri_info->prefix.rm_eo = 4;
                term->iri_info->path.rm_so = 4;
                term->iri_info->path.rm_eo = UUIDSTR_SIZE + 6;
                term->iri_info->frag.rm_so = -1;
                term->iri_info->frag.rm_eo = -1;
                term->iri_info->nsm = NULL;

            } else term->data = strdup (uuid_str);
        } else {
            log_error ("No data provided for term.");
            return LSUP_VALUE_ERR;
        }
    }

    if (term->type == LSUP_TERM_LT_LITERAL) {
        if (!metadata) {
            log_warn ("Lang tag is NULL. Creating a non-tagged literal.");
            term->type = LSUP_TERM_LITERAL;
        } else {
            char *lang_str = (char *) metadata;
            log_trace("Lang string: %s", lang_str);
            // Lang tags longer than 7 characters will be truncated.
            strncpy(term->lang, lang_str, sizeof (term->lang) - 1);
            if (strlen (term->lang) < 1) {
                log_error ("Lang tag cannot be an empty string.");
                return LSUP_VALUE_ERR;
            }
            term->lang[7] = '\0';
        }
    }

    if (term->type == LSUP_TERM_LITERAL) {
        term->datatype = metadata;
        if (! term->datatype) term->datatype = LSUP_default_datatype;
        log_trace ("Storing data type: %s", term->datatype->data);

        if (! LSUP_IS_IRI (term->datatype )) {
            log_error (
                    "Literal data tpe is not a IRI: %s",
                    term->datatype ->data);

            return LSUP_VALUE_ERR;
        }

        if (term->datatype != LSUP_default_datatype) {
            uint32_t dtype_hash = LSUP_term_hash (term->datatype );

            LSUP_Term *tmp = (LSUP_Term *) LSUP_tcache_get (dtype_hash);
            if (!tmp) LSUP_tcache_add (dtype_hash, term->datatype);
            else if (term->datatype != tmp) {
                free (term->datatype);
                term->datatype = tmp;
            }
        }

        log_trace ("Datatype address: %p", term->datatype);
        log_trace ("Datatype hash: %lu", LSUP_term_hash (term->datatype));

    } else if (term->type == LSUP_TERM_BNODE) {
        // TODO This is not usable for global skolemization.
        term->bnode_id = XXH64 (
                term->data, strlen (term->data) + 1, HASH_SEED);
    }

    return LSUP_OK;
}


LSUP_Key
LSUP_term_hash (const LSUP_Term *term)
{
    LSUP_Buffer *buf;

    if (UNLIKELY (!term)) buf = BUF_DUMMY;
    else buf = LSUP_term_serialize (term);

    LSUP_Key key = LSUP_buffer_hash (buf);

    LSUP_buffer_free (buf);

    return key;
}


/* DEPRECATED
bool LSUP_term_equals (const LSUP_Term *term1, const LSUP_Term *term2)
{
    LSUP_TermType type1, type2;
    char *data1, *data2;

    // Normalize IRI data before comparing.
    if (term1->type == LSUP_TERM_NS_IRIREF) {
        type1 = LSUP_TERM_IRIREF;
        if (UNLIKELY (LSUP_nsmap_normalize_uri (
            term1->iri_info->nsm, term1->data, &data1
        ) != LSUP_OK)) return LSUP_ERROR;
    } else {
        type1 = term1->type;
        data1 = term1->data;
    }

    if (term2->type == LSUP_TERM_NS_IRIREF) {
        type2 = LSUP_TERM_IRIREF;
        if (UNLIKELY (LSUP_nsmap_normalize_uri (
            term2->iri_info->nsm, term2->data, &data2
        ) != LSUP_OK)) return LSUP_ERROR;
    } else {
        type2 = term2->type;
        data2 = term2->data;
    }

    if (type1 != type2) return false;

    int cmp = strcmp (data1, data2);
    if (term1->type == LSUP_TERM_NS_IRIREF) free (data1);
    if (term2->type == LSUP_TERM_NS_IRIREF) free (data2);

    if (cmp != 0) return false;

    if (term1->type == LSUP_TERM_LITERAL)
        return term1->datatype == term2->datatype;

    if (term1->type == LSUP_TERM_LT_LITERAL)
        return strncmp (term1->lang, term2->lang, sizeof (term1->lang)) == 0;

    return true;
}
*/


void LSUP_term_done (LSUP_Term *term)
{
    free (term->data);
    term->data = NULL;
}


void LSUP_term_free (LSUP_Term *term)
{
    if (LIKELY (term != NULL)) {
        free (term->data);
        free (term);
    }
}


LSUP_NSMap *
LSUP_iriref_nsm (LSUP_Term *iri)
{
    if (iri->type != LSUP_TERM_IRIREF && iri->type != LSUP_TERM_NS_IRIREF) {
        log_error ("Term is not a IRI ref type.");
        return NULL;
    }

    return iri->iri_info->nsm;
}


char *
LSUP_iriref_prefix (LSUP_Term *iri)
{
    if (iri->type != LSUP_TERM_IRIREF && iri->type != LSUP_TERM_NS_IRIREF) {
        log_error ("Term is not a IRI ref type.");
        return NULL;
    }

    if (iri->iri_info->prefix.rm_so == -1) return NULL;

    size_t len = iri->iri_info->prefix.rm_eo - iri->iri_info->prefix.rm_so;

    return strndup (iri->data + iri->iri_info->prefix.rm_so, len);
}


char *
LSUP_iriref_path (LSUP_Term *iri)
{
    if (iri->type != LSUP_TERM_IRIREF && iri->type != LSUP_TERM_NS_IRIREF) {
        log_error ("Term is not a IRI ref type.");
        return NULL;
    }

    if (iri->iri_info->path.rm_so == -1) return NULL;

    size_t len = iri->iri_info->path.rm_eo - iri->iri_info->path.rm_so;

    return strndup (iri->data + iri->iri_info->path.rm_so, len);
}


char *
LSUP_iriref_frag (LSUP_Term *iri)
{
    if (iri->type != LSUP_TERM_IRIREF && iri->type != LSUP_TERM_NS_IRIREF) {
        log_error ("Term is not a IRI ref type.");
        return NULL;
    }

    if (iri->iri_info->frag.rm_so == -1) return NULL;

    size_t len = iri->iri_info->frag.rm_eo - iri->iri_info->frag.rm_so;

    return strndup (iri->data + iri->iri_info->frag.rm_so, len);
}


/*
 * Triple API.
 */

LSUP_Triple *
LSUP_triple_new(LSUP_Term *s, LSUP_Term *p, LSUP_Term *o)
{
    LSUP_Triple *spo = malloc (sizeof (*spo));
    if (!spo) return NULL;

    if (UNLIKELY (LSUP_triple_init (spo, s, p, o))) {
        free (spo);
        return NULL;
    }

    return spo;
}


LSUP_Triple *
LSUP_triple_new_from_btriple (const LSUP_BufferTriple *sspo)
{
    LSUP_Triple *spo = malloc (sizeof (*spo));
    if (!spo) return NULL;

    spo->s = LSUP_term_new_from_buffer (sspo->s);
    spo->p = LSUP_term_new_from_buffer (sspo->p);
    spo->o = LSUP_term_new_from_buffer (sspo->o);

    return spo;
}


LSUP_BufferTriple *
LSUP_triple_serialize (const LSUP_Triple *spo)
{
    LSUP_BufferTriple *sspo = malloc (sizeof (*sspo));
    if (!sspo) return NULL;

    sspo->s = LSUP_term_serialize (spo->s);
    sspo->p = LSUP_term_serialize (spo->p);
    sspo->o = LSUP_term_serialize (spo->o);

    return sspo;
}


LSUP_rc
LSUP_triple_init (LSUP_Triple *spo, LSUP_Term *s, LSUP_Term *p, LSUP_Term *o)
{
    // TODO validate term types.
    spo->s = s;
    spo->p = p;
    spo->o = o;

    return LSUP_OK;
}


void
LSUP_triple_done (LSUP_Triple *spo)
{
    if (UNLIKELY (!spo)) return;

    LSUP_term_free (spo->s);
    LSUP_term_free (spo->p);
    LSUP_term_free (spo->o);
}


void
LSUP_triple_free (LSUP_Triple *spo)
{
    if (UNLIKELY (!spo)) return;

    LSUP_term_free (spo->s);
    LSUP_term_free (spo->p);
    LSUP_term_free (spo->o);

    free (spo);
}


LSUP_rc
LSUP_tcache_add (const LSUP_Key key, LSUP_Term *term)
{
    struct term_cache_t *entry;

    HASH_FIND_INT (LSUP_term_cache, &key, entry);
    // Many calls will likely attempt inserting duplicates after the first one.
    if (LIKELY (entry)) return LSUP_NOACTION;

    MALLOC_GUARD (entry, LSUP_MEM_ERR);
    entry->key = key;
    entry->term = term;
    HASH_ADD_INT (LSUP_term_cache, key, entry);

    return LSUP_OK;
}


const LSUP_Term *
LSUP_tcache_get (const LSUP_Key key)
{
    struct term_cache_t *entry;

    HASH_FIND_INT (LSUP_term_cache, &key, entry);
    if (entry) log_trace ("ID found for key %u: %s", key, entry->term->data);
    else log_trace ("No ID found for key %u.", key);

    return (entry) ? entry->term : NULL;
}


/*
 * Extern inline functions.
 */

LSUP_Key LSUP_term_hash (const LSUP_Term *term);
LSUP_Term *LSUP_iriref_new (const char *data, LSUP_NSMap *nsm);
LSUP_Term *LSUP_literal_new (const char *data, LSUP_Term *datatype);
LSUP_Term *LSUP_lt_literal_new (const char *data, char *lang);
LSUP_Term *LSUP_bnode_new (const char *data);
bool LSUP_term_equals (const LSUP_Term *term1, const LSUP_Term *term2);
LSUP_Term *LSUP_triple_pos (const LSUP_Triple *trp, LSUP_TriplePos n);
LSUP_Key LSUP_triple_hash (const LSUP_Triple *trp);
